function [sigmar,righteig,psavec] = psa_subs(A,wA,epsln,pars)


% gammac is the lower bound for the second derivatives during a vertical search
if pars.steig == 1
	% standard eigenvalue problem
    gammac = 2;
else
	% not a standard eigenvalue problem
    gammac = 40000;
end

m = pars.m;
n = pars.n;
eigvals = pars.eigvals;
righteig = pars.righteig;
kmax = pars.kmax;


% normA is used in the termination criterion during vertical searches
if iscell(A)
	% for polynomials and nonlinear eigenvalues, use the norm of the leading coefficient
	normA = norm(A{pars.termindex});
else
	normA = norm(A);
end



%%%%%%%%%%%%%%%%%%%%%
%% compute eigenvalues and pseudospectrum
%%%%%%%%%%%%%%%%%%%%%
epslns = logspace(-10,log10(epsln),2);

nx = 150;
ny = 150;

if pars.isplot
	XL = linspace(pars.realbounds(1),pars.realbounds(2),nx);
	YL = linspace(pars.imagbounds(1),pars.imagbounds(2),ny);

	[X,Y,Z] = compute_ps_poly(A,wA,XL,YL,pars.steig,pars.funname);
    
	hold off;
	plot_ps_poly(eigvals,X,Y,Z,epslns);
	pause(0.1);
end

if pars.iskeyboard
	keyboard
end
%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% initializations for the main loop
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
warning off
[~,v] = ps(A,wA,righteig,pars.steig,pars.funname);
% initial subspace is 1-dimensional
Vk = v;
sigmaold = real(righteig);


if pars.isprint
	fprintf('initially, sigma %f\n',sigmaold);
end

k = 1;

%%%%%%%%%%%%%%%%%%%
% These are used for termination
%%%%%%%%%%%%%%%%%%%
% (1) A vertical search will be performed (i) if two consecutive local searches
%	  return the same up to a tolerance and (ii) if a minimum number of local searches
%	  has been performed since the last vertical search. The variable globalcheckmin
%	  corresponds to the minimum number of local searches so that a vertical search
%	  can be performed. The variable sinceglobalcheck represents the number of local
%	  searches since the last vertical search. Hence, (ii) holds if 
%					sinceglobalcheck >= globalcheckmin
% (2) The total number of local searches performed cannot exceed maxiter.
%%%%%%%%%%%%%%%%%%%
globalcheckmin = 7;
iter = 0;
sinceglobalcheck = 0;
maxiter = 1000;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


while iter <= maxiter
%%%%%%%%%%%%%%%%%%%%%%%
%% subspace iteration, main loop
%%%%%%%%%%%%%%%%%%%%%%%

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % pseudospectrum of projected matrix or matrix polynomial 
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	if iscell(A)
        for l = 1:m
            Ak{l} = A{l}*Vk;
        end
	else
		Ak{1} = A*Vk;
		Ak{2} = -Vk;
	end

	if pars.isplot
		eigvals = [];
		[~,~,Z] = compute_ps_poly(Ak,wA,XL,YL,pars.steig,pars.funname);
		plot_ps_poly(eigvals,X,Y,Z,epslns);
		title(['k = ',num2str(k)]);
	end
	%%%%%%%%%%%%%%%%%%%%%%%
	%%%%%%%%%%%%%%%%%%%%%%%


	pars.tol = max([1e-4*( ((1e-5-1)/(kmax-1))*(k-1)  +  1	)	epsln^2*5*10^-16	normA*5*10^-16	]);
	% require more accuracy as the dimension of subspace increases


    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% compute the pseudospectral abscissa for the projected matrix or matrix polynomial
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	funname = @fpsa;
	c = [1;0];
	pars.A = Ak;
	pars.wA = wA;
	pars.epsln = epsln;
	pars.x0 = [real(righteig);	imag(righteig)];
	pars.gamma = gammac;
	temp = pars.isprint;
	pars.isprint = 0;
	temp2 = pars.iskeyboard;
	pars.iskeyboard = 0;


	[sigmar,x] = eigopt_constrained(funname,c,pars);


	pars.isprint = temp;
	pars.iskeyboard = temp2;


	if (sigmar == inf)
		if k == 1
			psavec(1) = righteig;
		end
		righteig = inf;
		return;
	end

	righteig = sigmar + i*x(2);
	pars.sigma = sigmar;

	if pars.isprint
		fprintf('iter %d, subspace dim %d, sigma %.10f\n',iter+1,k,sigmar);
	end
	%%%%%%%%%%%%%%%%%%%%%%%%%
	%%%%%%%%%%%%%%%%%%%%%%%%%



	if ((sinceglobalcheck+1 >= globalcheckmin) && (abs(sigmar - sigmaold) < 10^-3))
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% check for global convergence
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		funname = @fdist_instab;
		if pars.steig == 1
            pars.A{1} = A;
            pars.A{2} = -eye(n);
		else
			pars.A = A;
		end
		bounds = struct('lb',pars.imagbounds(1),'ub',pars.imagbounds(2));
		pars.initial = x(2);
		pars.gamma = -4;
		pars.tol = normA*pars.termtol;
		temp = pars.isprint;
		pars.isprint = 0;
		temp2 = pars.iskeyboard;
		pars.iskeyboard = 0;

		[f2,omega,~] = eigopt(funname,bounds,pars);
		pars.isprint = temp;
		pars.iskeyboard = temp2;
		sinceglobalcheck = 0;
			

        %%%%%%%%%
		%%%%%%%%%
		% Terminate with convergence
		%%%%%%%%%
		if (f2 > epsln - pars.tol)
			psavec(iter+1) = righteig;

			if pars.isplot
				plot(real(righteig),imag(righteig),'*r');
				pause(0.1);
			end

				
			%%%%%%%%%%%%%%%%
			% Improve the final accuracy wih a final local search
			%%%%%%%%%%%%%%%%
			pars.A = Ak;

			funname = @fpsa;
			pars.wA = wA;
			pars.epsln = epsln;
			pars.x0 = [real(righteig);	imag(righteig)];
			pars.gamma = gammac;
			pars.tol = max([5*10^-11  epsln^2*5*10^-16  normA*5*10^-16]);

			if pars.isprint
				fprintf('*** Termination, Vertical Search Performed ***\n');
				fprintf('*** Improving accuracy before termination with convergence\n');
			end

			temp = pars.isprint;
			pars.isprint = 0;
			temp2 = pars.iskeyboard;
			pars.iskeyboard = 0;

			[sigmar,x] = eigopt_constrained(funname,c,pars);
			pars.isprint = temp;
			pars.iskeyboard = temp2;

			righteig = sigmar + i*x(2);
			%%%%%%%%%%%%%%%%
			%%%%%%%%%%%%%%%%


			sigmar = real(sigmar);
			return;

		else
		%%%%%%%%%
		%%%%%%%%%

            %%%%%%%%%%%%
			%%%%%%%%%%%%
			% Restart (better region in epsln-pseudospectrum is found)
			%%%%%%%%%%%%
			righteig = sigmar + i*omega;

			k = 1;
			[~,v] = ps(A,wA,righteig,pars.steig,pars.funname);
			Vk = v;
			psavec(iter+1) = righteig;

			if pars.isprint
				fprintf('*** Restarting, Vertical Search Performed ***\n');
			end

			if pars.isplot
				plot(real(righteig),imag(righteig),'*r');
				pause(0.1);
			end

			sigmaold = sigmar;
			iter = iter + 1;
			%%%%%%%%%%%%%%%%%
			%%%%%%%%%%%%%%%%%

		end

	else
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	% no check for global convergence
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

		if (k == kmax)

			%%%%%%%%%%%%
			% Restart (maximal subspace dimension is reached)
			%%%%%%%%%%%%
			k = 1;
			[~,v] = ps(A,wA,righteig,pars.steig,pars.funname);
			Vk = v;
			psavec(iter+1) = righteig;

			if pars.isprint
				fprintf('*** Restarting, Maximal Subspace Dimension Reached ***\n');
			end

			if pars.isplot
				plot(real(righteig),imag(righteig),'*r');
				pause(0.1);
			end

			sigmaold = sigmar;
			iter = iter + 1;
			sinceglobalcheck = sinceglobalcheck + 1;
			%%%%%%%%%%%%%%%%%
			%%%%%%%%%%%%%%%%%
		
		else
			%%%%%%%%%%%%%%%%%
			% expand subspace
			%%%%%%%%%%%%%%%%%
			k = k+1;
			[~,v] = ps(A,wA,righteig,pars.steig,pars.funname);
			Vk = [Vk v];
			[Vk,~] = qr(Vk,0);
			psavec(iter+1) = righteig;


			if pars.isplot
				plot(real(righteig),imag(righteig),'*r');
				pause(0.1);
			end

			sigmaold = sigmar;
			iter = iter+1;
			sinceglobalcheck = sinceglobalcheck + 1;
			%%%%%%%%%%%%%%%
			%%%%%%%%%%%%%%%
		end

	end

if pars.iskeyboard
	keyboard
end


%%%%%%%%%%%%%%%%%%
% end of main loop
%%%%%%%%%%%%%%%%%%
end


%%%%%%%%%
% Terminate without convergence
%%%%%%%%%
fprintf('*** Could not converge after %d iterations\n',maxiter);
sigmar = real(sigmar);
%%%%%%%%%
%%%%%%%%%

return;
