function [sigmar,righteig,psavec] = psa_subs_init(A,wA,epsln,pars)
%
% Call:	[sigmar,righteig,psavec] = psa_subs_init(A,wA,epsln,pars);
%
% Task: Computes the pseudospectral abscissa for matrices, matrix polynomials and
%       nonlinear eigenvalue problems
%
% The details are described in the following paper.
%		K. Meerbergen, E. Mengi, W. Michiels and R. Van Beeumen. Computation of Pseudospectral Abscissa
%		for Large Scale Nonlinear Eigenvalue Problems. Submitted to IMAJNA. 
%		Available at http://home.ku.edu.tr/~emengi/papers/papers.html
%
% See also the file "sample_calls".
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Input:
%
% A							nxn matrix (for standard psa), or
%							cell of matrices of size m (for polynomial and nonlinear psa), that is
%                           A{j+1} corresponds to A_j in (*).
%
% wA						weights, 1 for standard psa
%							otherwise a vector of nonnegative scalars (some of which are possibily inf) of size m
%
% epsln						positive real number
%
% pars						structure, possibly with the following fields
%							pars.kmax		--- maximal dimension of the subspace
%							pars.isprint	--- 0 do not print out the details (default), otherwise print
%							pars.isplot		--- 0 do not plot the pseudospectra, progress and iterates (default), otherwise plot
%							pars.iskeyboard --- 0 do not interact with the user (default), otherwise interact
%							pars.righteig	--- rightmost eigenvalue of the matrix-valued function (not essential to specify 
%												for polynomial or standard eigenvalue problem)
%							pars.funname	--- name of the function computing f_j(lambda) for j = 1,...,m and its derivatives in (*) below.
%													(*)     F(lambda)  =  f_0(lambda) A_0  +  f_1(lambda) A_1  +  ...  +  f_m(lambda) A_m
%												It is not necessary to provide this function for a polynomial or a standard eigenvalue problem
%												See the 'Note' below.
%							pars.realbounds (vector of size 2) 
%											--- an interval where the real part of a rightmost point
%												resides (if not provided, the interval [real(re)-15,real(re)+15]
%												is used, where re is the rightmost eigenvalue.)	
%							pars.imagbounds (vector of size 2) 
%											--- an interval where the imaginary part of a rightmost point
%												resides (if not provided, the interval [imag(re)-15,imag(re)+15]
%												is used, where re is the rightmost eigenvalue.)
%							pars.termindex	(integer in {0, 1, ..., m})
%											---	in case of a nonlinear eigenvalue problem (including matrix polynomials)
%												the index j of A_j in (*) that will determine the termination
%												after vertical searches. Termination will occur if the optimal
%												value 'opt' returned by a vertical search satisfies
%													(**)	opt <= epsln + pars.termtol*norm(A_j)
%												where pars.termtol is an input parameter as well. (default pars.termindex = m) 
%							pars.termtol	--- The termination parameter used after a vertical search. Termination
%												occurs if (**) holds. (default pars.termtol = 10^-6)
%
%		Note:	pars.funname takes one input parameter and must return three output parameters, specifically
%				[f,g,gr] = funname(z) where
%							f(j)	-	f_j(z) (f_j is as defined in (*))
%							g(j)	-	f_j'(z) (complex differentiation, e.g., (z^j)' = j*(z^{j-1}))
%							gr(j,:) -	partial derivatives of Re f_j(z), that is
%										gr(j,1) - d (Re f_j)(z)/ d Re(z)
%										gr(j,2) - d (Re f_j)(z) / d Im(z)
% See "delay.m" as an example.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Output:
%
% sigmar					computed epsln-pseudospectral abscissa
%
% righteig					computed globally rightmost point
%
% psavec					iterates of the algorithm leading to righteig
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialize
%%%%%%%%%%%%%%%%%%%%%%%%%

if iscell(A)
	[n,m] = size(A{1});
else
	[n,m] = size(A);
end


if isfield(pars, 'kmax')
	
	pars.kmax = min(pars.kmax,n);

else

	if n > 10
		pars.kmax = round(n/4);
	else
		pars.kmax = n;
	end
	
end


% default: do not print
if ~isfield(pars,'isprint')
	pars.isprint = 0;
end


% default: do not plot			
if ~isfield(pars,'isplot')
	pars.isplot = 0;
end

% default do not interact with the user
if ~isfield(pars,'iskeyboard')
	pars.iskeyboard = 0;
end

% default value of pars.termindex is m
% unless pars.termindex is provided explicitly		
if ~isfield(pars,'termindex')
	if iscell(A)
		pars.termindex = length(A);
	else
		pars.termindex = 1;
	end
end

% default value of pars.termtol is 10^-6
% unless pars.termtol is provided explicitly
if ~isfiled(pars,'termtol')
	pars.termtol = 10^-6;
end





%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% standard eigenvalue problem,  pars.steig = 1
% A - nxn, wA = 1
%
% polynomial eigenvalue problem, pars.steig = 0
% A - cell(m+1) for degree m polynomial s.t. A{j+1} = A_j for j = 0,...,m 
%
% other nonlinear eigenvalue problems, pars.steig = -1
% pars.funname must be provided and computes f_j(lambda) in 
%               F(lambda) = sum_{j=0}^m f_j(lambda) A_j also derivatives,
% A - cell(m+1) s.t. A{j+1} = A_j for j = 0,...,m
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if iscell(A)
	pars.m = length(A);
	pars.n = size(A{1},1);

	% not the standard eigenvalue problem
	if isfield(pars,'funname')
		pars.steig = -1; % nonlinear eigenvalue problem (not necessarily polynomial)
	else
		pars.steig = 0; % polynomial eigenvalue problem
	end
else
	pars.m = 1;
	pars.n = size(A,1);
	pars.steig = 1; % standard eigenvalue problem, needed by psa_sub.m
end
%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%
%% compute rightmost eigenvalue
%%%%%%%%%%%%%%%%%%%%%

if pars.steig >= 0
% polynomial or standard
    if ~isfield(pars,'righteig')
    % if rightmost eigenvalue is not specified
        eigvals = mypolyeig(A);
        [maxeig,mindex] = max(real(eigvals));
    else
    % if rightmost eigenvalue is specified
        eigvals = pars.righteig;
        mindex = 1;
    end
else
% nonlinear, rightmost eigenvalue must be specified
	if ~isfield(pars,'righteig')
		eigvals = 0;
		mindex = 1;
	else
		eigvals = pars.righteig;
		mindex = 1;
	end
end

pars.eigvals = eigvals;

[~,si] = max(real(eigvals));
pars.righteig = eigvals(si);
%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%
% More initializations
%%%%%%%%%%%%%%%%%%%

if ~isfield(pars,'realbounds')
	pars.realbounds(1) = real(pars.righteig)-15;
	pars.realbounds(2) = real(pars.righteig)+15;
end

if ~isfield(pars,'imagbounds')
	pars.imagbounds(1) = imag(pars.righteig)-15;
	pars.imagbounds(2) = imag(pars.righteig)+15;
end

if ~isfield(pars,'funname')
	pars.funname = 'dummy'; 
end


[sigmar,righteig,psavec] = psa_subs(A,wA,epsln,pars);

return;
